# -*- coding: utf-8 -*-
from spacy.matcher import Matcher
from spacy.tokens import Doc, Span
from phrase_detective.lib import merge
from phrase_detective.regx import REGX


class NounPhraseRecognizer:
  """Customerized component to detect noun phrases in ``spacy.tokens.Doc`` object. The ``NP`` values are stored in ``doc._.noun_phrases``,

  Attributes:
    ext_name (str): customized extension field name
    matcher (spacy.mathcer.Matcher): Rule maker for detecting ``NP``
  """

  def __init__(self, nlp):
    """Initialize the pipeline component. The shared nlp instance is used to initialize the matcher.
    
    Args:
      nlp (spacy.Language): language environment
    """

    self.ext_name = "noun_phrases"
    self.matcher = Matcher(nlp.vocab)
    np_patterns = REGX[nlp.meta["lang"]]["np"]
    self.matcher.add("NP", np_patterns)

    if not Doc.has_extension(self.ext_name):
      Doc.set_extension(self.ext_name, default=[])

  def __call__(self, doc):
    """Apply the pipeline component on a ``Doc`` object and modify it if matches are found.

    Returns:
      Doc: with customized extension ``doc._.noun_phrases``
    """

    matches = self.matcher(doc)

    phrases = []
    ranges = [(start, end) for _, start, end in matches]
    refined_matches = merge(ranges)
    for start, end in refined_matches:
      np = Span(doc, start, end)
      phrases.append(np)
    doc._.noun_phrases = phrases
    return doc




def get_verb_phrases(doc):
  """Get ``German`` verb phrases according to dependency parsing tags and verb detection

  Args:
    doc (spacy.tokens.Doc): sentence with meta information

  Returns:
    list: verb phrases with grammar description
  """

  verb_phrases = []
  for v in doc._.verbs:
    vp = {"lemma": v.lemma_, "complete verb": v.text, "direct object": [], "indirect object": []} 
    for obj in v.children:
      if obj.i == v.i - 1 and obj.dep_ == "aux" and v.text != v.lemma_ :
        vp["complete verb"] = " ".join([t.text for t in obj.subtree]) + " " + v.text 
      if obj.dep_ == "dobj":
        vp["direct object"].append(" ".join([t.text for t in obj.subtree]))
      if obj.dep_ == "dative":
        vp["indirect object"].append(" ".join([t.text for t in obj.subtree]))
    if len(vp["direct object"]) > 0 or len(vp["indirect object"]) > 0:
      verb_phrases.append(vp)
  return verb_phrases


de_rules = {
  "vp": get_verb_phrases,
}



# -*- coding: utf-8 -*-
from spacy.matcher import Matcher
from spacy.tokens import Doc, Span
from phrase_detective.lib import merge
from phrase_detective.regx import REGX


class PrepPhraseRecognizer:
  """Customerized component to detect noun phrases in ``spacy.tokens.Doc`` object. The ``PP`` values are stored in ``doc._.prep_phrases``.
  
  Attributes:
    ext_name (str): customized extension field name
    matcher (spacy.mathcer.Matcher): Rule maker for detecting ``PP``

  """

  def __init__(self, nlp):
    """Initialize the pipeline component. The shared nlp instance is used to initialize the matcher, which detects POS, TAG, DEP.

    Args:
      nlp (spacy.Language): language environment
    """

    self.ext_name = "prep_phrases"
    self.matcher = Matcher(nlp.vocab)
    pp_patterns = REGX[nlp.meta["lang"]]["pp"]
    self.matcher.add("PP", pp_patterns)

    if not Doc.has_extension(self.ext_name):
      Doc.set_extension(self.ext_name, default=[])
    

  def __call__(self, doc):
    """Apply the pipeline component on a Doc object and modify it if matches are found.
    
    Returns:
      Doc: with customized extension ``doc._.prep_phrases``
    """

    matches = self.matcher(doc)

    phrases = []
    ranges = [(start, end) for _, start, end in matches]
    refined_matches = merge(ranges)
    for start, end in refined_matches:
      if doc[start].text not in ['of', 'to', 'a']: # TODO: It is bad
        pp = Span(doc, start, end)
        phrases.append(pp)
    doc._.prep_phrases = phrases
    return doc




"""Define regular expression usage of ``spacy.matcher.Matcher`` for ``English``
"""

# NP noun phrase pattern
NP = [
  [{"POS": "DET"}, {"POS": "ADJ"}, {"POS": "NOUN"}],
  [{"POS": "DET", "OP": "?"}, {"TAG": "VBG"}, {"POS": "NOUN"}],
]

# PP preposition phrase pattern
PP = [
  [{"TAG": "IN"}, {"POS": "DET"}, {"POS": "NOUN", "DEP": "pobj"}],
  [{"TAG": "IN"}, {"TAG": "PRP", "DEP": "pobj"}],
  [{"TAG": "IN"}, {"TAG": "CD", "DEP": "pobj"}],
]

# VERB Knowledge pattern
VERB = [
  [{"POS": "VERB"}],
  [{"TAG": "VB"}],
  [{"TAG": "VBZ"}],
  [{"TAG": "VBN"}],
  [{"TAG": "VBP"}],
  [{"TAG": "VBG"}],
  [{"TAG": "VBD"}],
]

VERB_PASSIVE = [
  [{"DEP": "aux", "OP": "?"}, {"POS": "AUX", "DEP": "auxpass"}, {"POS": "VERB"}],
]

en_regx = {
  "verb": VERB,
  "verb_passive": VERB_PASSIVE,
  "pp": PP,
  "np": NP,
}



# -*- coding: utf-8 -*-
from spacy.matcher import Matcher
from spacy.tokens import Doc, Span, Token
from phrase_detective.lib import merge
from phrase_detective.regx import REGX
from phrase_detective.rules import RULES

class VerbKnowledgeRecognizer:
  """Customerized component to detect verb knowledge in ``spacy.tokens.Doc`` object. The corresponding values are stored in ``doc._.verbs``, ``doc._.passive_phrases``, and ``doc._.verb_phrases``.

  Attributes:
    ext_name (str): customized extension field name
    matcher (spacy.mathcer.Matcher): Rule maker for detecting ``VERB`` and ``PASSIVE``
  """

  def __init__(self, nlp):
    """Initialize the pipeline component. The shared nlp instance is used to initialize the matcher.
    
    Args:
      nlp (spacy.Language): language environment
    """

    self.nlp = nlp
    self.matcher = Matcher(nlp.vocab)

    lang = nlp.meta["lang"]
    patterns = REGX[lang] 
    verb_patterns = patterns["verb"]
    passive_phrase_patterns = patterns["verb_passive"]
    self.rules = RULES[lang]

    self.matcher.add("VERB", verb_patterns)
    self.matcher.add("PASSIVE", passive_phrase_patterns)

    if not Doc.has_extension("verbs"):
      Doc.set_extension("verbs", default=[])
    if not Doc.has_extension("passive_phrases"):
      Doc.set_extension("passive_phrases", default=[])
    if not Doc.has_extension("verb_phrases"):
      Doc.set_extension("verb_phrases", default=[])

    self._reset_matchstore()

  def _reset_matchstore(self):
    self.match_store = {
      "VERB": [],
      "PASSIVE": [],
    }

  def __call__(self, doc):
    """Apply the pipeline component on a Doc object and modify it if matches are found.

    Returns:
      Doc: with customized extensions ``doc._.verb_phrases``, ``doc._.passive_phrases``, `doc._.verbs` 
    """
    matches = self.matcher(doc)

    for match_id, start, end in matches:
      match_name = self.nlp.vocab.strings[match_id]
      self.match_store[match_name].append((start, end))

    self._get_verbs(doc, self.match_store["VERB"])
    self._get_passive_phrases(doc, self.match_store["PASSIVE"])
    self._get_verb_phrases(doc)

    self._reset_matchstore()

    return doc

  def _get_verbs(self, doc, matches):
    verbs = []
    refined_matches = merge(matches)
    for start, _ in refined_matches:
      v = doc[start]
      verbs.append(v)
    doc._.verbs = verbs 

  def _get_passive_phrases(self, doc, matches):
    passive_phrases = []
    refined_matches = merge(matches)
    for start, end in refined_matches:
      p = Span(doc, start, end)
      passive_phrases.append(p)
    doc._.passive_phrases = passive_phrases 

  def _get_verb_phrases(self, doc):
    doc._.verb_phrases = self.rules["vp"](doc)




"""Regular Expression used to analyze phrases

.. code:: python

  from phrase_detective.regx import REGX

  print(REGX)
"""

from .en import en_regx
from .es import es_regx
from .de import de_regx

REGX = {
  "en": en_regx,
  "es": es_regx,
  "de": de_regx,
}



# -*- coding: utf-8 -*-
"""This module ``phrase_detective.components`` implements ``spacy`` component for detecting noun phrase, prep phrase and verb phrase. 
"""

from .noun_phrases import NounPhraseRecognizer
from .prep_phrases import PrepPhraseRecognizer
from .verb_knowledge import VerbKnowledgeRecognizer



"""Define regular expression usage of ``spacy.matcher.Matcher`` for ``Spanish``
"""


# NP noun phrase pattern
NP = [
  [{"POS": "DET", "OP": "+"}, {"POS": "ADJ", "OP": "?"}, {"POS": "NOUN"}],
  [{"POS": "DET"}, {"POS": "NOUN"}, {"POS": "ADV", "OP": "?"}, {"POS": "ADJ"}],
  [{"POS": "NOUN"}, {"POS": "ADJ"}],
  [{"POS": "DET", "OP": "?"}, {"POS": "NUM"}, {"POS": "ADJ", "OP": "?"}, {"POS": "NOUN"}],
]

# PP preposition phrase pattern
PP = [
  [{"POS": "ADP"}, {"POS": "DET", "OP": "?"}, {"POS": "ADJ", "OP": "?"}, {"POS": "NUM", "OP": "?"}, {"POS": "NOUN"}],
  [{"POS": "ADP"}, {"POS": "DET", "OP": "?"}, {"POS": "NOUN"}, {"POS": "ADJ", "OP": "?"}],
  [{"POS": "ADP"}, {"POS": "DET", "OP": "?"}, {"POS": "PROPN"}],
  [{"POS": "ADP"}, {"POS": "PRON"}],
]

# VERB Knowledge pattern
VERB = [
  [{"POS": "VERB"}],
  [{"POS": "AUX"}],
]

VERB_PASSIVE = [
  [{"LEMMA": "ser"}, {"TAG": {"REGEX": "VERB_.*Tense=Past|VerbForm=Part"}}],
  [{"LOWER": "se", "DEP": "obj"}, {"TAG": {"REGEX": "VERB_.*"}}],
]


es_regx = {
  "verb": VERB,
  "verb_passive": VERB_PASSIVE,
  "pp": PP,
  "np": NP,
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['phrase_detective',
 'phrase_detective.components',
 'phrase_detective.regx',
 'phrase_detective.rules']

package_data = \
{'': ['*']}

install_requires = \
['spacy>=3.1.0,<4.0.0']

setup_kwargs = {
    'name': 'phrase-detective',
    'version': '0.1.16',
    'description': 'Phrase recognizer component for spacy pipeline',
    'long_description': '# Installation from pip3\n\n```shell\npip3 install --verbose phrase_detective \npython -m spacy download en_core_web_trf\npython -m spacy download es_dep_news_trf\npython -m spacy download de_dep_news_trf\n```\n\n# Usage\n\nPlease refer to [api docs](https://qishe-nlp.github.io/phrase-detective/)\n\n### Detect noun phrases \n```\nimport spacy\nfrom spacy import Language\nfrom phrase_detective import NounPhraseRecognizer, PKG_INDICES\n\n@Language.factory("nprecog")\ndef create_np_parser(nlp: Language, name: str):\n  return NounPhraseRecognizer(nlp) \n\ndef noun_phrase(lang, sentence):\n  nlp = spacy.load(PKG_INDICES[lang])\n  nlp.add_pipe("nprecog")\n  doc = nlp(sentence)\n  for np in doc._.noun_phrases:\n    print(np.text)\n\n```\n### Detect preposition phrases \n\n```\nimport spacy\nfrom spacy import Language\nfrom phrase_detective import PrepPhraseRecognizer, PKG_INDICES\n\n@Language.factory("pprecog")\ndef create_pp_parser(nlp: Language, name: str):\n  return PrepPhraseRecognizer(nlp) \n\ndef prep_phrase(lang, sentence):\n  nlp = spacy.load(PKG_INDICES[lang])\n  nlp.add_pipe("pprecog")\n  doc = nlp(sentence)\n  for np in doc._.prep_phrases:\n    print(np.text)\n```\n\n### Detect verb phrases \n\n```\nimport spacy\nfrom spacy import Language\nfrom phrase_detective import VerbKnowledgeRecognizer, PKG_INDICES\n\n@Language.factory("vkbrecog")\ndef create_vkb_parser(nlp: Language, name: str):\n  return VerbKnowledgeRecognizer(nlp) \n\ndef verb_knowledge(lang, sentence):\n  nlp = spacy.load(PKG_INDICES[lang])\n  nlp.add_pipe("vkbrecog")\n  doc = nlp(sentence)\n  for v in doc._.verbs:\n    print("TEXT: {}, TAG: {}, FORM: {}, ORIGNAL: {}".format(v.text, v.tag_, spacy.explain(v.tag_), v.lemma_))\n  for pp in doc._.passive_phrases:\n    print(pp.text)\n  for vp in doc._.verb_phrases:\n    print(vp)\n```\n\n# Development\n\n### Clone project\n```\ngit clone https://github.com/qishe-nlp/phrase-detective.git\n```\n\n### Install [poetry](https://python-poetry.org/docs/)\n\n### Install dependencies\n```\npoetry update\n```\n\n### Test and Issue\n```\npoetry run pytest -rP\n```\nwhich run tests under `tests/*`\n\n### Create sphix docs\n```\npoetry shell\ncd apidocs\nsphinx-apidoc -f -o source ../phrase_detective\nmake html\npython -m http.server -d build/html\n```\n\n### Hose docs on github pages\n```\ncp -rf apidocs/build/html/* docs/\n```\n\n### Build\n* Change `version` in `pyproject.toml` and `phrase_detective/__init__.py`\n* Build python package by `poetry build`\n\n### Git commit and push\n\n### Publish from local dev env\n* Set pypi test environment variables in poetry, refer to [poetry doc](https://python-poetry.org/docs/repositories/)\n* Publish to pypi test by `poetry publish -r test`\n\n### Publish through CI \n\n* Github action build and publish package to [test pypi repo](https://test.pypi.org/)\n\n```\ngit tag [x.x.x]\ngit push origin master\n```\n\n* Manually publish to [pypi repo](https://pypi.org/) through [github action](https://github.com/qishe-nlp/phrase-detective/actions/workflows/pypi.yml)\n\n',
    'author': 'Phoenix Grey',
    'author_email': 'phoenix.grey0108@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/qishe-nlp/phrase-detective',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
"""``phrase_detective`` Package

Components for ``spacy`` pipeline, to detect noun phrase, prep phrase, verb related knowledge

.. topic:: Install package

  .. code:: shell
    
    $ pip3 install --verbose phrase_detective
    $ python -m spacy download en_core_web_trf
    $ python -m spacy download es_dep_news_trf
    $ python -m spacy download de_dep_news_trf

.. topic:: Example: Detect noun phrases 

  .. code:: python

    import spacy
    from spacy import Language
    from phrase_detective import NounPhraseRecognizer, PKG_INDICES

    @Language.factory("nprecog")
    def create_np_parser(nlp: Language, name: str):
      return NounPhraseRecognizer(nlp) 

    def noun_phrase(lang, sentence):
      nlp = spacy.load(PKG_INDICES[lang])
      nlp.add_pipe("nprecog")
      doc = nlp(sentence)
      for np in doc._.noun_phrases:
        print(np.text)

.. topic:: Example: Detect preposition phrases 

  .. code:: python 

    import spacy
    from spacy import Language
    from phrase_detective import PrepPhraseRecognizer, PKG_INDICES

    @Language.factory("pprecog")
    def create_pp_parser(nlp: Language, name: str):
      return PrepPhraseRecognizer(nlp) 

    def prep_phrase(lang, sentence):
      nlp = spacy.load(PKG_INDICES[lang])
      nlp.add_pipe("pprecog")
      doc = nlp(sentence)
      for np in doc._.prep_phrases:
        print(np.text)

.. topic:: Example: Detect verb phrases 

  .. code:: python

    import spacy
    from spacy import Language
    from phrase_detective import VerbKnowledgeRecognizer, PKG_INDICES

    @Language.factory("vkbrecog")
    def create_vkb_parser(nlp: Language, name: str):
      return VerbKnowledgeRecognizer(nlp) 

    def verb_knowledge(lang, sentence):
      nlp = spacy.load(PKG_INDICES[lang])
      nlp.add_pipe("vkbrecog")
      doc = nlp(sentence)
      for v in doc._.verbs:
        print("TEXT: {}, TAG: {}, FORM: {}, ORIGNAL: {}".format(v.text, v.tag_, spacy.explain(v.tag_), v.lemma_))
      for pp in doc._.passive_phrases:
        print(pp.text)
      for vp in doc._.verb_phrases:
        print(vp)

"""

__version__ = '0.1.16'

from .components import NounPhraseRecognizer, PrepPhraseRecognizer, VerbKnowledgeRecognizer
from .constants import PKG_INDICES
